@@ -18,6 +18,7 @@
 package org.apache.drill.exec.physical.impl.mergereceiver;
 
 import java.io.IOException;
+import java.util.Arrays;
 import java.util.Comparator;
 import java.util.Iterator;
 import java.util.List;
@@ -41,6 +42,7 @@ import org.apache.drill.exec.expr.ClassGenerator.HoldingContainer;
 import org.apache.drill.exec.expr.CodeGenerator;
 import org.apache.drill.exec.expr.ExpressionTreeMaterializer;
 import org.apache.drill.exec.expr.fn.FunctionGenerationHelper;
+import org.apache.drill.exec.ops.ExchangeFragmentContext;
 import org.apache.drill.exec.ops.FragmentContext;
 import org.apache.drill.exec.ops.MetricDef;
 import org.apache.drill.exec.physical.MinorFragmentEndpoint;
@@ -77,8 +79,8 @@ import org.apache.drill.exec.vector.CopyUtil;
 import org.apache.drill.exec.vector.FixedWidthVector;
 import org.apache.drill.exec.vector.ValueVector;
 
-import com.google.common.base.Preconditions;
-import com.google.common.collect.Lists;
+import org.apache.drill.shaded.guava.com.google.common.base.Preconditions;
+import org.apache.drill.shaded.guava.com.google.common.collect.Lists;
 import com.sun.codemodel.JConditional;
 import com.sun.codemodel.JExpr;
 
@@ -88,7 +90,7 @@ import io.netty.buffer.ByteBuf;
 /**
  * The MergingRecordBatch merges pre-sorted record batches from remote senders.
  */
-public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP> {
+public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP> implements RecordBatch {
   private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(MergingRecordBatch.class);
   private static final ControlsInjector injector = ControlsInjectorFactory.getInjector(MergingRecordBatch.class);
 
@@ -96,7 +98,7 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
 
   private RecordBatchLoader[] batchLoaders;
   private final RawFragmentBatchProvider[] fragProviders;
-  private final FragmentContext context;
+  private final ExchangeFragmentContext context;
   private VectorContainer outgoingContainer;
   private MergingReceiverGeneratorBase merger;
   private final MergingReceiverPOP config;
@@ -109,12 +111,11 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
   private RawFragmentBatch[] incomingBatches;
   private int[] batchOffsets;
   private PriorityQueue <Node> pqueue;
-  private RawFragmentBatch emptyBatch = null;
   private RawFragmentBatch[] tempBatchHolder;
   private long[] inputCounts;
   private long[] outputCounts;
 
-  public static enum Metric implements MetricDef{
+  public enum Metric implements MetricDef {
     BYTES_RECEIVED,
     NUM_SENDERS,
     NEXT_WAIT_NANOS;
@@ -125,7 +126,7 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
     }
   }
 
-  public MergingRecordBatch(final FragmentContext context,
+  public MergingRecordBatch(final ExchangeFragmentContext context,
                             final MergingReceiverPOP config,
                             final RawFragmentBatchProvider[] fragProviders) throws OutOfMemoryException {
     super(config, context, true, context.newOperatorContext(config));
@@ -136,9 +137,11 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
     this.config = config;
     this.inputCounts = new long[config.getNumSenders()];
     this.outputCounts = new long[config.getNumSenders()];
+
+    // Register this operator's buffer allocator so that incoming buffers are owned by this allocator
+    context.getBuffers().getCollector(config.getOppositeMajorFragmentId()).setAllocator(oContext.getAllocator());
   }
 
-  @SuppressWarnings("resource")
   private RawFragmentBatch getNext(final int providerIndex) throws IOException {
     stats.startWait();
     final RawFragmentBatchProvider provider = fragProviders[providerIndex];
@@ -210,11 +213,11 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
           try {
             rawBatch = getNext(p);
           } catch (final IOException e) {
-            context.fail(e);
+            context.getExecutorState().fail(e);
             return IterOutcome.STOP;
           }
         }
-        if (rawBatch == null && !context.shouldContinue()) {
+        if (rawBatch == null && !context.getExecutorState().shouldContinue()) {
           clearBatches(rawBatches);
           return IterOutcome.STOP;
         }
@@ -241,12 +244,12 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
             while ((rawBatch = getNext(p)) != null && rawBatch.getHeader().getDef().getRecordCount() == 0) {
               // Do nothing
             }
-            if (rawBatch == null && !context.shouldContinue()) {
+            if (rawBatch == null && !context.getExecutorState().shouldContinue()) {
               clearBatches(rawBatches);
               return IterOutcome.STOP;
             }
           } catch (final IOException e) {
-            context.fail(e);
+            context.getExecutorState().fail(e);
             clearBatches(rawBatches);
             return IterOutcome.STOP;
           }
@@ -315,7 +318,7 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
           // SchemaChangeException, so check/clean catch clause below.
         } catch(final SchemaChangeException e) {
           logger.error("MergingReceiver failed to load record batch from remote host.  {}", e);
-          context.fail(e);
+          context.getExecutorState().fail(e);
           return IterOutcome.STOP;
         }
         batch.release();
@@ -328,7 +331,7 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
       // Ensure all the incoming batches have the identical schema.
       // Note: RecordBatchLoader permutes the columns to obtain the same columns order for all batches.
       if (!isSameSchemaAmongBatches(batchLoaders)) {
-        context.fail(new SchemaChangeException("Incoming batches for merging receiver have different schemas!"));
+        context.getExecutorState().fail(new SchemaChangeException("Incoming batches for merging receiver have different schemas!"));
         return IterOutcome.STOP;
       }
 
@@ -351,7 +354,7 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
         merger = createMerger();
       } catch (final SchemaChangeException e) {
         logger.error("Failed to generate code for MergingReceiver.  {}", e);
-        context.fail(e);
+        context.getExecutorState().fail(e);
         return IterOutcome.STOP;
       }
 
@@ -380,12 +383,12 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
             } else {
               batchLoaders[b].clear();
               batchLoaders[b] = null;
-              if (!context.shouldContinue()) {
+              if (!context.getExecutorState().shouldContinue()) {
                 return IterOutcome.STOP;
               }
             }
           } catch (IOException | SchemaChangeException e) {
-            context.fail(e);
+            context.getExecutorState().fail(e);
             return IterOutcome.STOP;
           }
         }
@@ -418,11 +421,11 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
 
           assert nextBatch != null || inputCounts[node.batchId] == outputCounts[node.batchId]
               : String.format("Stream %d input count: %d output count %d", node.batchId, inputCounts[node.batchId], outputCounts[node.batchId]);
-          if (nextBatch == null && !context.shouldContinue()) {
+          if (nextBatch == null && !context.getExecutorState().shouldContinue()) {
             return IterOutcome.STOP;
           }
         } catch (final IOException e) {
-          context.fail(e);
+          context.getExecutorState().fail(e);
           return IterOutcome.STOP;
         }
 
@@ -456,7 +459,7 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
           // TODO:  Clean:  DRILL-2933:  That load(...) no longer throws
           // SchemaChangeException, so check/clean catch clause below.
         } catch(final SchemaChangeException ex) {
-          context.fail(ex);
+          context.getExecutorState().fail(ex);
           return IterOutcome.STOP;
         }
         incomingBatches[node.batchId].release();
@@ -532,7 +535,10 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
 
   @Override
   public BatchSchema getSchema() {
-    return outgoingContainer.getSchema();
+    if (outgoingContainer.hasSchema()) {
+      return outgoingContainer.getSchema();
+    }
+    return null;
   }
 
   @Override
@@ -548,7 +554,7 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
         }
         final RawFragmentBatch batch = getNext(i);
         if (batch == null) {
-          if (!context.shouldContinue()) {
+          if (!context.getExecutorState().shouldContinue()) {
             state = BatchState.STOP;
           } else {
             state = BatchState.DONE;
@@ -561,7 +567,6 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
         }
         tempBatchHolder[i] = batch;
         for (final SerializedField field : batch.getHeader().getDef().getFieldList()) {
-          @SuppressWarnings("resource")
           final ValueVector v = outgoingContainer.addOrGet(MaterializedField.create(field));
           v.allocateNew();
         }
@@ -578,6 +583,19 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
     return outgoingPosition;
   }
 
+  @Override
+  public void kill(final boolean sendUpstream) {
+    if (sendUpstream) {
+      informSenders();
+    } else {
+      close();
+    }
+
+    for (final RawFragmentBatchProvider provider : fragProviders) {
+      provider.kill(context);
+    }
+  }
+
   private void informSenders() {
     logger.info("Informing senders of request to terminate sending.");
     final FragmentHandle handlePrototype = FragmentHandle.newBuilder()
@@ -592,7 +610,9 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
               .setReceiver(context.getHandle())
               .setSender(sender)
               .build();
-      context.getControlTunnel(providingEndpoint.getEndpoint()).informReceiverFinished(new OutcomeListener(), finishedReceiver);
+      context.getController()
+        .getTunnel(providingEndpoint.getEndpoint())
+        .informReceiverFinished(new OutcomeListener(), finishedReceiver);
     }
   }
 
@@ -611,25 +631,17 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
 
     @Override
     public void interrupted(final InterruptedException e) {
-      if (context.shouldContinue()) {
+      if (context.getExecutorState().shouldContinue()) {
         final String errMsg = "Received an interrupt RPC outcome while sending ReceiverFinished message";
         logger.error(errMsg, e);
-        context.fail(new RpcException(errMsg, e));
+        context.getExecutorState().fail(new RpcException(errMsg, e));
       }
     }
   }
 
   @Override
   protected void killIncoming(final boolean sendUpstream) {
-    if (sendUpstream) {
-      informSenders();
-    }
-    this.hasMoreIncoming = false;
-    close();
-
-    for (final RawFragmentBatchProvider provider : fragProviders) {
-      provider.kill(context);
-    }
+    //No op
   }
 
   @Override
@@ -678,7 +690,6 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
 
   private void allocateOutgoing() {
     for (final VectorWrapper<?> w : outgoingContainer) {
-      @SuppressWarnings("resource")
       final ValueVector v = w.getValueVector();
       if (v instanceof FixedWidthVector) {
         AllocationHelper.allocate(v, OUTGOING_BATCH_SIZE, 1);
@@ -728,11 +739,11 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
     }
   }
 
-  public final MappingSet MAIN_MAPPING = new MappingSet( (String) null, null, ClassGenerator.DEFAULT_SCALAR_MAP, ClassGenerator.DEFAULT_SCALAR_MAP);
-  public final MappingSet LEFT_MAPPING = new MappingSet("leftIndex", null, ClassGenerator.DEFAULT_SCALAR_MAP, ClassGenerator.DEFAULT_SCALAR_MAP);
-  public final MappingSet RIGHT_MAPPING = new MappingSet("rightIndex", null, ClassGenerator.DEFAULT_SCALAR_MAP, ClassGenerator.DEFAULT_SCALAR_MAP);
-  GeneratorMapping COPIER_MAPPING = new GeneratorMapping("doSetup", "doCopy", null, null);
-  public final MappingSet COPIER_MAPPING_SET = new MappingSet(COPIER_MAPPING, COPIER_MAPPING);
+  private final MappingSet MAIN_MAPPING = new MappingSet( (String) null, null, ClassGenerator.DEFAULT_SCALAR_MAP, ClassGenerator.DEFAULT_SCALAR_MAP);
+  private final MappingSet LEFT_MAPPING = new MappingSet("leftIndex", null, ClassGenerator.DEFAULT_SCALAR_MAP, ClassGenerator.DEFAULT_SCALAR_MAP);
+  private final MappingSet RIGHT_MAPPING = new MappingSet("rightIndex", null, ClassGenerator.DEFAULT_SCALAR_MAP, ClassGenerator.DEFAULT_SCALAR_MAP);
+  private GeneratorMapping COPIER_MAPPING = new GeneratorMapping("doSetup", "doCopy", null, null);
+  private final MappingSet COPIER_MAPPING_SET = new MappingSet(COPIER_MAPPING, COPIER_MAPPING);
 
   private void generateComparisons(final ClassGenerator<?> g, final VectorAccessible batch) throws SchemaChangeException {
     g.setMappingSet(MAIN_MAPPING);
@@ -817,4 +828,11 @@ public class MergingRecordBatch extends AbstractRecordBatch<MergingReceiverPOP>
     super.close();
   }
 
+  @Override
+  public void dump() {
+    logger.error("MergingRecordBatch[container={}, outgoingPosition={}, incomingBatches={}, batchOffsets={}, "
+            + "tempBatchHolder={}, inputCounts={}, outputCounts={}]",
+        container, outgoingPosition, Arrays.toString(incomingBatches), Arrays.toString(batchOffsets),
+        Arrays.toString(tempBatchHolder), Arrays.toString(inputCounts), Arrays.toString(outputCounts));
+  }
 }
