@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -6,9 +6,9 @@
  * to you under the Apache License, Version 2.0 (the
  * "License"); you may not use this file except in compliance
  * with the License.  You may obtain a copy of the License at
- * <p/>
+ *
  * http://www.apache.org/licenses/LICENSE-2.0
- * <p/>
+ *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
@@ -17,20 +17,16 @@
  */
 package org.apache.drill.exec.planner.index;
 
-import com.google.common.base.Charsets;
-import com.google.common.base.Preconditions;
-import com.google.common.collect.Lists;
-
-import com.google.common.collect.Maps;
 import org.apache.calcite.plan.RelOptUtil;
 import org.apache.calcite.rel.RelNode;
 import org.apache.calcite.rel.metadata.RelMdUtil;
 import org.apache.calcite.rel.type.RelDataType;
 import org.apache.calcite.rex.RexBuilder;
 import org.apache.calcite.rex.RexCall;
+import org.apache.calcite.rex.RexInputRef;
 import org.apache.calcite.rex.RexLiteral;
 import org.apache.calcite.rex.RexNode;
-import org.apache.calcite.rex.RexInputRef;
+import org.apache.calcite.rex.RexUtil;
 import org.apache.calcite.sql.SqlKind;
 import org.apache.calcite.sql.fun.SqlStdOperatorTable;
 import org.apache.calcite.sql.type.SqlTypeName;
@@ -39,7 +35,6 @@ import org.apache.drill.common.expression.ExpressionStringBuilder;
 import org.apache.drill.common.expression.LogicalExpression;
 import org.apache.drill.exec.physical.base.DbGroupScan;
 import org.apache.drill.exec.physical.base.GroupScan;
-import org.apache.drill.exec.planner.common.DrillRelOptUtil;
 import org.apache.drill.exec.planner.common.DrillScanRelBase;
 import org.apache.drill.exec.planner.logical.DrillOptiq;
 import org.apache.drill.exec.planner.logical.DrillParseContext;
@@ -53,13 +48,17 @@ import org.apache.hadoop.hbase.HConstants;
 import org.ojai.store.QueryCondition;
 
 import java.io.UnsupportedEncodingException;
-import java.util.Arrays;
 import java.util.ArrayList;
-import java.util.Iterator;
-import java.util.List;
+import java.util.Arrays;
 import java.util.HashMap;
 import java.util.HashSet;
+import java.util.Iterator;
+import java.util.List;
 import java.util.Map;
+import org.apache.drill.shaded.guava.com.google.common.base.Charsets;
+import org.apache.drill.shaded.guava.com.google.common.base.Preconditions;
+import org.apache.drill.shaded.guava.com.google.common.collect.Lists;
+import org.apache.drill.shaded.guava.com.google.common.collect.Maps;
 
 public class MapRDBStatistics implements Statistics {
   static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(MapRDBStatistics.class);
@@ -135,8 +134,17 @@ public class MapRDBStatistics implements Statistics {
     Map<String, StatisticsPayload> payloadMap;
     if ((scanRel instanceof DrillScanRel && ((DrillScanRel)scanRel).getGroupScan() instanceof DbGroupScan)
         || (scanRel instanceof ScanPrel && ((ScanPrel)scanRel).getGroupScan() instanceof DbGroupScan)) {
-      if (condition == null && fullTableScanPayload != null) {
-        return fullTableScanPayload.getRowCount();
+      if (condition == null) {
+        if (tabIdxName != null) {
+          // fiStatsCache also stores the total rowcount for the index. Complex indexes usually would have different
+          // rowcount than the primary table since they are stored in a different manner.
+          StatisticsPayload fiPayloadMap = fIStatsCache.get(tabIdxName);
+          if (fiPayloadMap != null) {
+            return fiPayloadMap.getRowCount();
+          }
+        } else if (fullTableScanPayload != null) {
+          return fullTableScanPayload.getRowCount();
+        }
       } else if (condition != null) {
         conditionAsStr = convertRexToString(condition, scanRel.getRowType());
         payloadMap = statsCache.get(conditionAsStr);
@@ -193,9 +201,17 @@ public class MapRDBStatistics implements Statistics {
           return anyPayload.getRowCount();
         }
       }
-    } else if (condition == null
-        && fullTableScanPayload != null) {
-      return fullTableScanPayload.getRowCount();
+    } else if (condition == null) {
+      if (tabIdxName != null) {
+        // fiStatsCache also stores the total rowcount for the index. Complex indexes usually would have different
+        // rowcount than the primary table since they are stored in a different manner.
+        StatisticsPayload fiPayloadMap = fIStatsCache.get(tabIdxName);
+        if (fiPayloadMap != null) {
+          return fiPayloadMap.getRowCount();
+        }
+      } else if (fullTableScanPayload != null) {
+        return fullTableScanPayload.getRowCount();
+      }
     }
     if (condition != null) {
       conditionAsStr = condition.toString();
@@ -317,11 +333,11 @@ public class MapRDBStatistics implements Statistics {
     // Get the stats payload for full table (has total rows in the table)
     StatisticsPayload ftsPayload = jTabGrpScan.getFirstKeyEstimatedStats(null, null, scanRel);
     addToCache(null, null, context, ftsPayload, jTabGrpScan, scanRel, scanRel.getRowType());
-    addToCache(null, jTabGrpScan.getAverageRowSizeStats(null), ftsPayload);
+    addToCache(null, jTabGrpScan.getFilterIndependentStats(null), ftsPayload);
     // Get the stats for all indexes
     for (IndexDescriptor idx: indexes) {
       StatisticsPayload idxPayload = jTabGrpScan.getFirstKeyEstimatedStats(null, idx, scanRel);
-      StatisticsPayload idxRowSizePayload = jTabGrpScan.getAverageRowSizeStats(idx);
+      StatisticsPayload idxRowSizePayload = jTabGrpScan.getFilterIndependentStats(idx);
       RelDataType newRowType;
       FunctionalIndexInfo functionInfo = idx.getFunctionalInfo();
       if (functionInfo.hasFunctional()) {
@@ -379,12 +395,12 @@ public class MapRDBStatistics implements Statistics {
     StatisticsPayload ftsPayload = jTabGrpScan.getFirstKeyEstimatedStats(null, null, scanRel);
 
     // Get the average row size for table and all indexes
-    addToCache(null, jTabGrpScan.getAverageRowSizeStats(null), ftsPayload);
+    addToCache(null, jTabGrpScan.getFilterIndependentStats(null), ftsPayload);
     if (ftsPayload == null || ftsPayload.getRowCount() == 0) {
       return;
     }
     for (IndexDescriptor idx : indexes) {
-      StatisticsPayload idxRowSizePayload = jTabGrpScan.getAverageRowSizeStats(idx);
+      StatisticsPayload idxRowSizePayload = jTabGrpScan.getFilterIndependentStats(idx);
       addToCache(idx, idxRowSizePayload, ftsPayload);
     }
 
@@ -419,9 +435,9 @@ public class MapRDBStatistics implements Statistics {
           convertToLogicalExpression(preProcIdxCondition, newRowType, settings, builder));
       // Cap rows/size at total rows in case of issues with DB APIs
       StatisticsPayload idxPayload = jTabGrpScan.getFirstKeyEstimatedStats(queryCondition, idx, scanRel);
-      double rowCount = Math.min(idxPayload.getRowCount(), ftsPayload.getRowCount());
-      double leadingRowCount = Math.min(idxPayload.getLeadingRowCount(), rowCount);
-      double avgRowSize = Math.min(idxPayload.getAvgRowSize(), ftsPayload.getAvgRowSize());
+      double rowCount = idxPayload.getRowCount();
+      double leadingRowCount = idxPayload.getLeadingRowCount();
+      double avgRowSize = idxPayload.getAvgRowSize();
       StatisticsPayload payload = new MapRDBStatisticsPayload(rowCount, leadingRowCount, avgRowSize);
       addToCache(idxCondition, idx, context, payload, jTabGrpScan, scanRel, newRowType);
       addBaseConditions(idxCondition, payload, false, baseConditionMap, scanRel.getRowType());
@@ -692,13 +708,13 @@ public class MapRDBStatistics implements Statistics {
       for(RexNode pred : RelOptUtil.conjunctions(condition)) {
         conditions.add(convertToStatsCondition(pred, index, context, scanRel, typesToProcess));
       }
-      return DrillRelOptUtil.composeConjunction(builder, conditions, false);
+      return RexUtil.composeConjunction(builder, conditions, false);
     } else if (condition.getKind() == SqlKind.OR) {
       final List<RexNode> conditions = Lists.newArrayList();
       for(RexNode pred : RelOptUtil.disjunctions(condition)) {
         conditions.add(convertToStatsCondition(pred, index, context, scanRel, typesToProcess));
       }
-      return DrillRelOptUtil.composeDisjunction(builder, conditions, false);
+      return RexUtil.composeDisjunction(builder, conditions, false);
     } else if (condition instanceof RexCall) {
       // LIKE operator - convert to a RANGE predicate, if possible
       if (typesToProcess.contains(SqlKind.LIKE)
@@ -834,7 +850,7 @@ public class MapRDBStatistics implements Statistics {
           startKey = prefix.getBytes(Charsets.UTF_8);
           stopKey = startKey.clone();
           boolean isMaxVal = true;
-          for (int i = stopKey.length - 1; i >= 0 ; --i) {
+          for (int i = stopKey.length - 1; i >= 0; --i) {
             int nextByteValue = (0xff & stopKey[i]) + 1;
             if (nextByteValue < 0xff) {
               stopKey[i] = (byte) nextByteValue;
@@ -940,7 +956,7 @@ public class MapRDBStatistics implements Statistics {
    * the given collection would contain only one index for each distinct row key in the collection
    */
   private IndexCollection distinctFKeyIndexes(IndexCollection indexes, RelNode scanRel) {
-    IndexCollection distinctIdxCollection = new DrillIndexCollection(scanRel, new HashSet<DrillIndexDescriptor>());
+    IndexCollection distinctIdxCollection = new MapRDBIndexCollection(scanRel, new HashSet<DrillIndexDescriptor>());
     Iterator<IndexDescriptor> iterator = indexes.iterator();
     Map<String, List<IndexDescriptor>> firstColIndexMap = new HashMap<>();
     while (iterator.hasNext()) {
