@@ -17,14 +17,14 @@
  */
 package org.apache.drill.exec.planner.sql;
 
-import com.google.common.collect.ImmutableMap;
-import com.google.common.collect.Lists;
+import org.apache.calcite.sql.fun.SqlStdOperatorTable;
+import org.apache.drill.shaded.guava.com.google.common.collect.ImmutableMap;
+import org.apache.drill.shaded.guava.com.google.common.collect.ImmutableSet;
+import org.apache.drill.shaded.guava.com.google.common.collect.Lists;
 
 import org.apache.calcite.avatica.util.TimeUnit;
 import org.apache.calcite.rel.type.RelDataType;
 import org.apache.calcite.rel.type.RelDataTypeFactory;
-import org.apache.calcite.runtime.CalciteException;
-import org.apache.calcite.runtime.Resources;
 import org.apache.calcite.sql.SqlCallBinding;
 import org.apache.calcite.sql.SqlCharStringLiteral;
 import org.apache.calcite.sql.SqlDynamicParam;
@@ -34,12 +34,11 @@ import org.apache.calcite.sql.SqlLiteral;
 import org.apache.calcite.sql.SqlNode;
 import org.apache.calcite.sql.SqlNumericLiteral;
 import org.apache.calcite.sql.SqlOperatorBinding;
-import org.apache.calcite.sql.fun.SqlAvgAggFunction;
 import org.apache.calcite.sql.parser.SqlParserPos;
 import org.apache.calcite.sql.type.SqlReturnTypeInference;
+import org.apache.calcite.sql.type.SqlTypeFamily;
 import org.apache.calcite.sql.type.SqlTypeName;
 
-import org.apache.calcite.sql.validate.SqlValidatorException;
 import org.apache.drill.common.expression.ExpressionPosition;
 import org.apache.drill.common.expression.FunctionCall;
 import org.apache.drill.common.expression.FunctionCallFactory;
@@ -48,15 +47,17 @@ import org.apache.drill.common.expression.MajorTypeInLogicalExpression;
 import org.apache.drill.common.exceptions.UserException;
 import org.apache.drill.common.types.TypeProtos;
 import org.apache.drill.common.types.Types;
+import org.apache.drill.exec.expr.annotations.FunctionTemplate;
 import org.apache.drill.exec.expr.fn.DrillFuncHolder;
-import org.apache.drill.exec.planner.physical.PlannerSettings;
+import org.apache.drill.exec.planner.types.DrillRelDataTypeSystem;
 import org.apache.drill.exec.resolver.FunctionResolver;
 import org.apache.drill.exec.resolver.FunctionResolverFactory;
 import org.apache.drill.exec.resolver.TypeCastRules;
-import org.apache.drill.exec.server.options.OptionManager;
 
 import java.util.List;
+import java.util.Set;
 
+@SuppressWarnings("WeakerAccess")
 public class TypeInferenceUtils {
   private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(TypeInferenceUtils.class);
 
@@ -74,11 +75,12 @@ public class TypeInferenceUtils {
       .put(TypeProtos.MinorType.DECIMAL18, SqlTypeName.DECIMAL)
       .put(TypeProtos.MinorType.DECIMAL28SPARSE, SqlTypeName.DECIMAL)
       .put(TypeProtos.MinorType.DECIMAL38SPARSE, SqlTypeName.DECIMAL)
+      .put(TypeProtos.MinorType.VARDECIMAL, SqlTypeName.DECIMAL)
       .put(TypeProtos.MinorType.TIME, SqlTypeName.TIME)
       .put(TypeProtos.MinorType.TIMESTAMP, SqlTypeName.TIMESTAMP)
       .put(TypeProtos.MinorType.VARBINARY, SqlTypeName.VARBINARY)
       .put(TypeProtos.MinorType.INTERVALYEAR, SqlTypeName.INTERVAL_YEAR_MONTH)
-      .put(TypeProtos.MinorType.INTERVALDAY, SqlTypeName.INTERVAL_DAY_TIME)
+      .put(TypeProtos.MinorType.INTERVALDAY, SqlTypeName.INTERVAL_DAY)
       .put(TypeProtos.MinorType.MAP, SqlTypeName.MAP)
       .put(TypeProtos.MinorType.LIST, SqlTypeName.ARRAY)
       .put(TypeProtos.MinorType.LATE, SqlTypeName.ANY)
@@ -102,8 +104,20 @@ public class TypeInferenceUtils {
       .put(SqlTypeName.TIME, TypeProtos.MinorType.TIME)
       .put(SqlTypeName.TIMESTAMP, TypeProtos.MinorType.TIMESTAMP)
       .put(SqlTypeName.VARBINARY, TypeProtos.MinorType.VARBINARY)
+      .put(SqlTypeName.INTERVAL_YEAR, TypeProtos.MinorType.INTERVALYEAR)
       .put(SqlTypeName.INTERVAL_YEAR_MONTH, TypeProtos.MinorType.INTERVALYEAR)
-      .put(SqlTypeName.INTERVAL_DAY_TIME, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_MONTH, TypeProtos.MinorType.INTERVALYEAR)
+      .put(SqlTypeName.INTERVAL_DAY, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_DAY_HOUR, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_DAY_MINUTE, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_DAY_SECOND, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_HOUR, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_HOUR_MINUTE, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_HOUR_SECOND, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_MINUTE, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_MINUTE_SECOND, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.INTERVAL_SECOND, TypeProtos.MinorType.INTERVALDAY)
+      .put(SqlTypeName.DECIMAL, TypeProtos.MinorType.VARDECIMAL)
 
       // SqlTypeName.CHAR is the type for Literals in Calcite, Drill treats Literals as VARCHAR also
       .put(SqlTypeName.CHAR, TypeProtos.MinorType.VARCHAR)
@@ -127,29 +141,31 @@ public class TypeInferenceUtils {
 
   private static final ImmutableMap<String, SqlReturnTypeInference> funcNameToInference = ImmutableMap.<String, SqlReturnTypeInference> builder()
       .put("DATE_PART", DrillDatePartSqlReturnTypeInference.INSTANCE)
-      .put("COUNT", DrillCountSqlReturnTypeInference.INSTANCE)
+      .put(SqlStdOperatorTable.TIMESTAMP_ADD.getName(), DrillTimestampAddTypeInference.INSTANCE)
+      .put(SqlKind.SUM.name(), DrillSumSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.COUNT.name(), DrillCountSqlReturnTypeInference.INSTANCE)
       .put("CONCAT", DrillConcatSqlReturnTypeInference.INSTANCE_CONCAT)
       .put("CONCATOPERATOR", DrillConcatSqlReturnTypeInference.INSTANCE_CONCAT_OP)
       .put("LENGTH", DrillLengthSqlReturnTypeInference.INSTANCE)
       .put("LPAD", DrillPadSqlReturnTypeInference.INSTANCE)
       .put("RPAD", DrillPadSqlReturnTypeInference.INSTANCE)
-      .put("LTRIM", DrillTrimSqlReturnTypeInference.INSTANCE)
-      .put("RTRIM", DrillTrimSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.LTRIM.name(), DrillTrimSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.RTRIM.name(), DrillTrimSqlReturnTypeInference.INSTANCE)
       .put("BTRIM", DrillTrimSqlReturnTypeInference.INSTANCE)
-      .put("TRIM", DrillTrimSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.TRIM.name(), DrillTrimSqlReturnTypeInference.INSTANCE)
       .put("CONVERT_TO", DrillConvertToSqlReturnTypeInference.INSTANCE)
-      .put("EXTRACT", DrillExtractSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.EXTRACT.name(), DrillExtractSqlReturnTypeInference.INSTANCE)
       .put("SQRT", DrillSqrtSqlReturnTypeInference.INSTANCE)
-      .put("CAST", DrillCastSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.CAST.name(), DrillCastSqlReturnTypeInference.INSTANCE)
       .put("FLATTEN", DrillDeferToExecSqlReturnTypeInference.INSTANCE)
       .put("KVGEN", DrillDeferToExecSqlReturnTypeInference.INSTANCE)
       .put("CONVERT_FROM", DrillDeferToExecSqlReturnTypeInference.INSTANCE)
 
       // Functions that return the same type
-      .put("LOWER", DrillSameSqlReturnTypeInference.INSTANCE)
-      .put("UPPER", DrillSameSqlReturnTypeInference.INSTANCE)
-      .put("INITCAP", DrillSameSqlReturnTypeInference.INSTANCE)
-      .put("REVERSE", DrillSameSqlReturnTypeInference.INSTANCE)
+      .put("LOWER", DrillSameSqlReturnTypeInference.THE_SAME_RETURN_TYPE)
+      .put("UPPER", DrillSameSqlReturnTypeInference.THE_SAME_RETURN_TYPE)
+      .put("INITCAP", DrillSameSqlReturnTypeInference.THE_SAME_RETURN_TYPE)
+      .put("REVERSE", DrillSameSqlReturnTypeInference.THE_SAME_RETURN_TYPE)
 
       // Window Functions
       // RANKING
@@ -160,18 +176,34 @@ public class TypeInferenceUtils {
       .put(SqlKind.ROW_NUMBER.name(), DrillRankingSqlReturnTypeInference.INSTANCE_BIGINT)
 
       // NTILE
-      .put("NTILE", DrillNTILESqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.NTILE.name(), DrillNTILESqlReturnTypeInference.INSTANCE)
+
+      // LEAD, LAG
+      .put(SqlKind.LEAD.name(), DrillLeadLagSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.LAG.name(), DrillLeadLagSqlReturnTypeInference.INSTANCE)
 
       // FIRST_VALUE, LAST_VALUE
-      .put("FIRST_VALUE", DrillSameSqlReturnTypeInference.INSTANCE)
-      .put("LAST_VALUE", DrillSameSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.FIRST_VALUE.name(), DrillSameSqlReturnTypeInference.THE_SAME_RETURN_TYPE)
+      .put(SqlKind.LAST_VALUE.name(), DrillSameSqlReturnTypeInference.THE_SAME_RETURN_TYPE)
 
       // Functions rely on DrillReduceAggregatesRule for expression simplification as opposed to getting evaluated directly
-      .put(SqlAvgAggFunction.Subtype.AVG.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
-      .put(SqlAvgAggFunction.Subtype.STDDEV_POP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
-      .put(SqlAvgAggFunction.Subtype.STDDEV_SAMP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
-      .put(SqlAvgAggFunction.Subtype.VAR_POP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
-      .put(SqlAvgAggFunction.Subtype.VAR_SAMP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.AVG.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.STDDEV_POP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.STDDEV_SAMP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.VAR_POP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.VAR_SAMP.name(), DrillAvgAggSqlReturnTypeInference.INSTANCE)
+      .put(SqlKind.MIN.name(), DrillSameSqlReturnTypeInference.ALL_NULLABLE)
+      .put(SqlKind.MAX.name(), DrillSameSqlReturnTypeInference.ALL_NULLABLE)
+      .build();
+
+  /**
+   * Set of the decimal functions which return type cannot be determined exactly for some reasons at the current stage.
+   * For example functions which takes as an parameter scale or precision of return type.
+   */
+  private static final Set<String> SET_SCALE_DECIMAL_FUNCTIONS = ImmutableSet.<String> builder()
+      .add("ROUND")
+      .add("TRUNC")
+      .add("TRUNCATE")
       .build();
 
   /**
@@ -193,6 +225,39 @@ public class TypeInferenceUtils {
     return getDrillTypeFromCalciteType(sqlTypeName);
   }
 
+  /**
+   * Returns {@link TypeProtos.MajorType} instance which corresponds to specified {@code RelDataType relDataType}
+   * with its nullability, scale and precision if it is available.
+   *
+   * @param relDataType RelDataType to convert
+   * @return {@link TypeProtos.MajorType} instance
+   */
+  public static TypeProtos.MajorType getDrillMajorTypeFromCalciteType(RelDataType relDataType) {
+    final SqlTypeName sqlTypeName = relDataType.getSqlTypeName();
+
+    TypeProtos.MinorType minorType = getDrillTypeFromCalciteType(sqlTypeName);
+    TypeProtos.MajorType.Builder typeBuilder = TypeProtos.MajorType.newBuilder().setMinorType(minorType);
+    switch (minorType) {
+      case VAR16CHAR:
+      case VARCHAR:
+      case VARBINARY:
+      case TIMESTAMP:
+        if (relDataType.getPrecision() > 0) {
+          typeBuilder.setPrecision(relDataType.getPrecision());
+        }
+        break;
+      case VARDECIMAL:
+        typeBuilder.setPrecision(relDataType.getPrecision());
+        typeBuilder.setScale(relDataType.getScale());
+    }
+    if (relDataType.isNullable()) {
+      typeBuilder.setMode(TypeProtos.DataMode.OPTIONAL);
+    } else {
+      typeBuilder.setMode(TypeProtos.DataMode.REQUIRED);
+    }
+    return typeBuilder.build();
+  }
+
   /**
    * Given a Calcite's SqlTypeName, return a Drill's corresponding TypeProtos.MinorType
    */
@@ -209,21 +274,13 @@ public class TypeInferenceUtils {
    */
   public static SqlReturnTypeInference getDrillSqlReturnTypeInference(
       final String name,
-      final List<DrillFuncHolder> functions,
-      final OptionManager optionManager) {
+      final List<DrillFuncHolder> functions) {
+
     final String nameCap = name.toUpperCase();
     if(funcNameToInference.containsKey(nameCap)) {
       return funcNameToInference.get(nameCap);
     } else {
-      switch (name) {
-        case "SUM":
-          return new DrillSumSqlReturnTypeInference(optionManager);
-        case "LEAD":
-        case "LAG":
-          return new DrillLeadLagSqlReturnTypeInference(optionManager);
-        default:
-          return new DrillDefaultSqlReturnTypeInference(functions, optionManager);
-      }
+      return new DrillDefaultSqlReturnTypeInference(functions);
     }
   }
 
@@ -239,11 +296,9 @@ public class TypeInferenceUtils {
 
   private static class DrillDefaultSqlReturnTypeInference implements SqlReturnTypeInference {
     private final List<DrillFuncHolder> functions;
-    private final OptionManager optionManager;
 
-    public DrillDefaultSqlReturnTypeInference(List<DrillFuncHolder> functions, OptionManager optionManager) {
+    public DrillDefaultSqlReturnTypeInference(List<DrillFuncHolder> functions) {
       this.functions = functions;
-      this.optionManager = optionManager;
     }
 
     @Override
@@ -255,12 +310,6 @@ public class TypeInferenceUtils {
             true);
       }
 
-      // Only the operand types are useful in the type inference.
-      // Also, since Decimal literals are treated as DOUBLE in Drill's exection,
-      // inference algorithm treats them as DOUBLE also
-      if(!optionManager.getOption(PlannerSettings.ENABLE_DECIMAL_DATA_TYPE_KEY).bool_val) {
-        opBinding = convertDecimalLiteralToDouble(opBinding);
-      }
       // The following logic is just a safe play:
       // Even if any of the input arguments has ANY type,
       // it "might" still be possible to determine the return type based on other non-ANY types
@@ -270,32 +319,47 @@ public class TypeInferenceUtils {
           // In summary, if we have a boolean output function in the WHERE-CLAUSE,
           // this logic can validate and execute user queries seamlessly
           boolean allBooleanOutput = true;
+          boolean isNullable = false;
           for (DrillFuncHolder function : functions) {
             if (function.getReturnType().getMinorType() != TypeProtos.MinorType.BIT) {
               allBooleanOutput = false;
               break;
             }
+            if (function.getReturnType().getMode() == TypeProtos.DataMode.OPTIONAL
+                || function.getNullHandling() == FunctionTemplate.NullHandling.NULL_IF_NULL) {
+              isNullable = true;
+            }
           }
 
-          if(allBooleanOutput) {
+          if (allBooleanOutput) {
             return factory.createTypeWithNullability(
-                factory.createSqlType(SqlTypeName.BOOLEAN), true);
+                factory.createSqlType(SqlTypeName.BOOLEAN), isNullable);
           } else {
             return factory.createTypeWithNullability(
                 factory.createSqlType(SqlTypeName.ANY),
                 true);
           }
+        } else if (SET_SCALE_DECIMAL_FUNCTIONS.contains(opBinding.getOperator().getName())
+            && getDrillTypeFromCalciteType(type) == TypeProtos.MinorType.VARDECIMAL) {
+          return factory.createTypeWithNullability(
+              factory.createSqlType(SqlTypeName.ANY),
+              true);
         }
       }
 
-      final DrillFuncHolder func = resolveDrillFuncHolder(opBinding, functions);
-      final RelDataType returnType = getReturnType(opBinding, func);
+      final FunctionCall functionCall = convertSqlOperatorBindingToFunctionCall(opBinding);
+      final DrillFuncHolder func = resolveDrillFuncHolder(opBinding, functions, functionCall);
+
+      final RelDataType returnType = getReturnType(opBinding,
+          func.getReturnType(functionCall.args), func.getNullHandling());
+
       return returnType.getSqlTypeName() == SqlTypeName.VARBINARY
           ? createCalciteTypeWithNullability(factory, SqlTypeName.ANY, returnType.isNullable())
               : returnType;
     }
 
-    private static RelDataType getReturnType(final SqlOperatorBinding opBinding, final DrillFuncHolder func) {
+    private static RelDataType getReturnType(final SqlOperatorBinding opBinding,
+        final TypeProtos.MajorType returnType, FunctionTemplate.NullHandling nullHandling) {
       final RelDataTypeFactory factory = opBinding.getTypeFactory();
 
       // least restrictive type (nullable ANY type)
@@ -303,7 +367,6 @@ public class TypeInferenceUtils {
           factory.createSqlType(SqlTypeName.ANY),
           true);
 
-      final TypeProtos.MajorType returnType = func.getReturnType();
       if (UNKNOWN_TYPE.equals(returnType)) {
         return nullableAnyType;
       }
@@ -322,7 +385,7 @@ public class TypeInferenceUtils {
           break;
 
         case REQUIRED:
-          switch (func.getNullHandling()) {
+          switch (nullHandling) {
             case INTERNAL:
               isNullable = false;
               break;
@@ -347,10 +410,7 @@ public class TypeInferenceUtils {
           throw new UnsupportedOperationException();
       }
 
-      return createCalciteTypeWithNullability(
-          factory,
-          sqlTypeName,
-          isNullable);
+      return convertToCalciteType(factory, returnType, isNullable);
     }
   }
 
@@ -367,10 +427,7 @@ public class TypeInferenceUtils {
   }
 
   private static class DrillSumSqlReturnTypeInference implements SqlReturnTypeInference {
-    private final OptionManager optionManager;
-    public DrillSumSqlReturnTypeInference(OptionManager optionManager) {
-      this.optionManager = optionManager;
-    }
+    private static final DrillSumSqlReturnTypeInference INSTANCE = new DrillSumSqlReturnTypeInference();
 
     @Override
     public RelDataType inferReturnType(SqlOperatorBinding opBinding) {
@@ -381,13 +438,6 @@ public class TypeInferenceUtils {
       final boolean isNullable = opBinding.getGroupCount() == 0
           || opBinding.getOperandType(0).isNullable();
 
-      // Only the operand types are useful in the type inference.
-      // Also, since Decimal literals are treated as DOUBLE in Drill's exection,
-      // inference algorithm treats them as DOUBLE also
-      if(!optionManager.getOption(PlannerSettings.ENABLE_DECIMAL_DATA_TYPE_KEY).bool_val) {
-        opBinding = convertDecimalLiteralToDouble(opBinding);
-      }
-
       if(getDrillTypeFromCalciteType(opBinding.getOperandType(0)) == TypeProtos.MinorType.LATE) {
         return createCalciteTypeWithNullability(
             factory,
@@ -395,15 +445,36 @@ public class TypeInferenceUtils {
             isNullable);
       }
 
+      // Determines SqlTypeName of the result.
+      // For the case when input may be implicitly casted to BIGINT, the type of result is BIGINT.
+      // Else for the case when input may be implicitly casted to FLOAT4, the type of result is DOUBLE.
+      // Else for the case when input may be implicitly casted to VARDECIMAL, the type of result is DECIMAL
+      // with the same scale as input and max allowed numeric precision.
+      // Else for the case when input may be implicitly casted to FLOAT8, the type of result is DOUBLE.
+      // When none of these conditions is satisfied, error is thrown.
+      // This order of checks is caused by the order of types in ResolverTypePrecedence.precedenceMap
       final RelDataType operandType = opBinding.getOperandType(0);
       final TypeProtos.MinorType inputMinorType = getDrillTypeFromCalciteType(operandType);
-      if(TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.BIGINT))
+      if (TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.BIGINT))
           == TypeProtos.MinorType.BIGINT) {
         return createCalciteTypeWithNullability(
             factory,
             SqlTypeName.BIGINT,
             isNullable);
-      } else if(TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.FLOAT8))
+      } else if (TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.FLOAT4))
+          == TypeProtos.MinorType.FLOAT4) {
+        return createCalciteTypeWithNullability(
+            factory,
+            SqlTypeName.DOUBLE,
+            isNullable);
+      } else if (TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.VARDECIMAL))
+          == TypeProtos.MinorType.VARDECIMAL) {
+        RelDataType sqlType = factory.createSqlType(SqlTypeName.DECIMAL,
+          DrillRelDataTypeSystem.DRILL_REL_DATATYPE_SYSTEM.getMaxNumericPrecision(),
+          Math.min(operandType.getScale(),
+            DrillRelDataTypeSystem.DRILL_REL_DATATYPE_SYSTEM.getMaxNumericScale()));
+        return factory.createTypeWithNullability(sqlType, isNullable);
+      } else if (TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.FLOAT8))
           == TypeProtos.MinorType.FLOAT8) {
         return createCalciteTypeWithNullability(
             factory,
@@ -519,6 +590,64 @@ public class TypeInferenceUtils {
     }
   }
 
+  private static class DrillTimestampAddTypeInference implements SqlReturnTypeInference {
+    private static final SqlReturnTypeInference INSTANCE = new DrillTimestampAddTypeInference();
+
+    @Override
+    public RelDataType inferReturnType(SqlOperatorBinding opBinding) {
+      RelDataTypeFactory factory = opBinding.getTypeFactory();
+      // operands count ond order is checked at parsing stage
+      RelDataType inputType = opBinding.getOperandType(2);
+      boolean isNullable = inputType.isNullable() || opBinding.getOperandType(1).isNullable();
+
+      SqlTypeName inputTypeName = inputType.getSqlTypeName();
+
+      TimeUnit qualifier = ((SqlLiteral) ((SqlCallBinding) opBinding).operand(0)).getValueAs(TimeUnit.class);
+
+      SqlTypeName sqlTypeName;
+
+      // follow up with type inference of reduced expression
+      switch (qualifier) {
+        case DAY:
+        case WEEK:
+        case MONTH:
+        case QUARTER:
+        case YEAR:
+        case NANOSECOND:  // NANOSECOND is not supported by Calcite SqlTimestampAddFunction.
+                          // Once it is fixed, NANOSECOND should be moved to the group below.
+          sqlTypeName = inputTypeName;
+          break;
+        case MICROSECOND:
+        case MILLISECOND:
+          // precision should be specified for MICROSECOND and MILLISECOND
+          return factory.createTypeWithNullability(
+              factory.createSqlType(SqlTypeName.TIMESTAMP, 3),
+              isNullable);
+        case SECOND:
+        case MINUTE:
+        case HOUR:
+          if (inputTypeName == SqlTypeName.TIME) {
+            sqlTypeName = SqlTypeName.TIME;
+          } else {
+            sqlTypeName = SqlTypeName.TIMESTAMP;
+          }
+          break;
+        default:
+          sqlTypeName = SqlTypeName.ANY;
+      }
+
+      // preserves precision of input type if it was specified
+      if (inputType.getSqlTypeName().allowsPrecNoScale()) {
+        RelDataType type = factory.createSqlType(sqlTypeName, inputType.getPrecision());
+        return factory.createTypeWithNullability(type, isNullable);
+      }
+      return createCalciteTypeWithNullability(
+          opBinding.getTypeFactory(),
+          sqlTypeName,
+          isNullable);
+    }
+  }
+
   private static class DrillSubstringSqlReturnTypeInference implements SqlReturnTypeInference {
     private static final DrillSubstringSqlReturnTypeInference INSTANCE = new DrillSubstringSqlReturnTypeInference();
 
@@ -616,6 +745,7 @@ public class TypeInferenceUtils {
     private static final DrillCastSqlReturnTypeInference INSTANCE = new DrillCastSqlReturnTypeInference();
 
     @Override
+    @SuppressWarnings("deprecation")
     public RelDataType inferReturnType(SqlOperatorBinding opBinding) {
       final RelDataTypeFactory factory = opBinding.getTypeFactory();
       final boolean isNullable = opBinding
@@ -674,29 +804,29 @@ public class TypeInferenceUtils {
   }
 
   private static class DrillLeadLagSqlReturnTypeInference implements SqlReturnTypeInference {
-    private final OptionManager optionManager;
-    public DrillLeadLagSqlReturnTypeInference(OptionManager optionManager) {
-      this.optionManager = optionManager;
-    }
-
+    private static final DrillLeadLagSqlReturnTypeInference INSTANCE = new DrillLeadLagSqlReturnTypeInference();
     @Override
     public RelDataType inferReturnType(SqlOperatorBinding opBinding) {
-      // Only the operand types are useful in the type inference.
-      // Also, since Decimal literals are treated as DOUBLE in Drill's exection,
-      // inference algorithm treats them as DOUBLE also
-      if(!optionManager.getOption(PlannerSettings.ENABLE_DECIMAL_DATA_TYPE_KEY).bool_val) {
-        opBinding = convertDecimalLiteralToDouble(opBinding);
-      }
-
       return opBinding.getTypeFactory().createTypeWithNullability(opBinding.getOperandType(0), true);
     }
   }
 
   private static class DrillSameSqlReturnTypeInference implements SqlReturnTypeInference {
-    private static final DrillSameSqlReturnTypeInference INSTANCE = new DrillSameSqlReturnTypeInference();
+    private static final DrillSameSqlReturnTypeInference THE_SAME_RETURN_TYPE = new DrillSameSqlReturnTypeInference(true);
+    private static final DrillSameSqlReturnTypeInference ALL_NULLABLE = new DrillSameSqlReturnTypeInference(false);
+
+    private final boolean preserveNullability;
+
+    public DrillSameSqlReturnTypeInference(boolean preserveNullability) {
+      this.preserveNullability = preserveNullability;
+    }
+
     @Override
     public RelDataType inferReturnType(SqlOperatorBinding opBinding) {
-      return opBinding.getOperandType(0);
+      if (preserveNullability) {
+        return opBinding.getOperandType(0);
+      }
+      return opBinding.getTypeFactory().createTypeWithNullability(opBinding.getOperandType(0), true);
     }
   }
 
@@ -704,27 +834,72 @@ public class TypeInferenceUtils {
     private static final DrillAvgAggSqlReturnTypeInference INSTANCE = new DrillAvgAggSqlReturnTypeInference();
     @Override
     public RelDataType inferReturnType(SqlOperatorBinding opBinding) {
-      final boolean isNullable = opBinding.getGroupCount() == 0 || opBinding.hasFilter() || opBinding.getOperandType(0).isNullable();
-      return createCalciteTypeWithNullability(
-          opBinding.getTypeFactory(),
-          SqlTypeName.DOUBLE,
-          isNullable);
+      final RelDataTypeFactory factory = opBinding.getTypeFactory();
+      // If there is group-by and the imput type is Non-nullable,
+      // the output is Non-nullable;
+      // Otherwise, the output is nullable.
+      final boolean isNullable = opBinding.getGroupCount() == 0
+          || opBinding.getOperandType(0).isNullable();
+
+      if (getDrillTypeFromCalciteType(opBinding.getOperandType(0)) == TypeProtos.MinorType.LATE) {
+        return createCalciteTypeWithNullability(
+            factory,
+            SqlTypeName.ANY,
+            isNullable);
+      }
+
+      // Determines SqlTypeName of the result.
+      // For the case when input may be implicitly casted to FLOAT4, the type of result is DOUBLE.
+      // Else for the case when input may be implicitly casted to VARDECIMAL, the type of result is DECIMAL
+      // with scale max(6, input) and max allowed numeric precision.
+      // Else for the case when input may be implicitly casted to FLOAT8, the type of result is DOUBLE.
+      // When none of these conditions is satisfied, error is thrown.
+      // This order of checks is caused by the order of types in ResolverTypePrecedence.precedenceMap
+      final RelDataType operandType = opBinding.getOperandType(0);
+      final TypeProtos.MinorType inputMinorType = getDrillTypeFromCalciteType(operandType);
+      if (TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.FLOAT4))
+          == TypeProtos.MinorType.FLOAT4) {
+        return createCalciteTypeWithNullability(
+            factory,
+            SqlTypeName.DOUBLE,
+            isNullable);
+      } else if (TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.VARDECIMAL))
+          == TypeProtos.MinorType.VARDECIMAL) {
+        RelDataType sqlType = factory.createSqlType(SqlTypeName.DECIMAL,
+            DrillRelDataTypeSystem.DRILL_REL_DATATYPE_SYSTEM.getMaxNumericPrecision(),
+            Math.min(Math.max(6, operandType.getScale()),
+                DrillRelDataTypeSystem.DRILL_REL_DATATYPE_SYSTEM.getMaxNumericScale()));
+        return factory.createTypeWithNullability(sqlType, isNullable);
+      } else if (TypeCastRules.getLeastRestrictiveType(Lists.newArrayList(inputMinorType, TypeProtos.MinorType.FLOAT8))
+          == TypeProtos.MinorType.FLOAT8) {
+        return createCalciteTypeWithNullability(
+            factory,
+            SqlTypeName.DOUBLE,
+            isNullable);
+      } else {
+        throw UserException
+            .functionError()
+            .message(String.format("%s does not support operand types (%s)",
+                opBinding.getOperator().getName(),
+                opBinding.getOperandType(0).getSqlTypeName()))
+            .build(logger);
+      }
     }
   }
 
-  private static DrillFuncHolder resolveDrillFuncHolder(final SqlOperatorBinding opBinding, final List<DrillFuncHolder> functions) {
-    final FunctionCall functionCall = convertSqlOperatorBindingToFunctionCall(opBinding);
+  private static DrillFuncHolder resolveDrillFuncHolder(final SqlOperatorBinding opBinding,
+      final List<DrillFuncHolder> functions, FunctionCall functionCall) {
     final FunctionResolver functionResolver = FunctionResolverFactory.getResolver(functionCall);
     final DrillFuncHolder func = functionResolver.getBestMatch(functions, functionCall);
 
     // Throw an exception
     // if no DrillFuncHolder matched for the given list of operand types
-    if(func == null) {
-      String operandTypes = "";
-      for(int i = 0; i < opBinding.getOperandCount(); ++i) {
-        operandTypes += opBinding.getOperandType(i).getSqlTypeName();
-        if(i < opBinding.getOperandCount() - 1) {
-          operandTypes += ",";
+    if (func == null) {
+      StringBuilder operandTypes = new StringBuilder();
+      for (int i = 0; i < opBinding.getOperandCount(); ++i) {
+        operandTypes.append(opBinding.getOperandType(i).getSqlTypeName());
+        if (i < opBinding.getOperandCount() - 1) {
+          operandTypes.append(",");
         }
       }
 
@@ -732,7 +907,7 @@ public class TypeInferenceUtils {
           .functionError()
           .message(String.format("%s does not support operand types (%s)",
               opBinding.getOperator().getName(),
-              operandTypes))
+              operandTypes.toString()))
           .build(logger);
     }
     return func;
@@ -741,15 +916,16 @@ public class TypeInferenceUtils {
   /**
    * For Extract and date_part functions, infer the return types based on timeUnit
    */
-  public static SqlTypeName getSqlTypeNameForTimeUnit(String timeUnit) {
-    switch (timeUnit.toUpperCase()){
-      case "YEAR":
-      case "MONTH":
-      case "DAY":
-      case "HOUR":
-      case "MINUTE":
+  public static SqlTypeName getSqlTypeNameForTimeUnit(String timeUnitStr) {
+    TimeUnit timeUnit = TimeUnit.valueOf(timeUnitStr);
+    switch (timeUnit) {
+      case YEAR:
+      case MONTH:
+      case DAY:
+      case HOUR:
+      case MINUTE:
         return SqlTypeName.BIGINT;
-      case "SECOND":
+      case SECOND:
         return SqlTypeName.DOUBLE;
       default:
         throw UserException
@@ -771,13 +947,13 @@ public class TypeInferenceUtils {
                                                              SqlTypeName sqlTypeName,
                                                              boolean isNullable) {
     RelDataType type;
-    if (sqlTypeName == SqlTypeName.INTERVAL_DAY_TIME) {
+    if (sqlTypeName.getFamily() == SqlTypeFamily.INTERVAL_DAY_TIME) {
       type = typeFactory.createSqlIntervalType(
           new SqlIntervalQualifier(
               TimeUnit.DAY,
               TimeUnit.MINUTE,
               SqlParserPos.ZERO));
-    } else if (sqlTypeName == SqlTypeName.INTERVAL_YEAR_MONTH) {
+    } else if (sqlTypeName.getFamily() == SqlTypeFamily.INTERVAL_YEAR_MONTH) {
       type = typeFactory.createSqlIntervalType(
           new SqlIntervalQualifier(
               TimeUnit.YEAR,
@@ -791,6 +967,25 @@ public class TypeInferenceUtils {
     return typeFactory.createTypeWithNullability(type, isNullable);
   }
 
+  /**
+   * Creates a RelDataType using specified RelDataTypeFactory which corresponds to specified TypeProtos.MajorType.
+   *
+   * @param typeFactory RelDataTypeFactory used to create the RelDataType
+   * @param drillType   the given TypeProtos.MajorType
+   * @param isNullable  nullability of the resulting type
+   * @return RelDataType which corresponds to specified TypeProtos.MajorType
+   */
+  public static RelDataType convertToCalciteType(RelDataTypeFactory typeFactory,
+                                                 TypeProtos.MajorType drillType, boolean isNullable) {
+    SqlTypeName sqlTypeName = getCalciteTypeFromDrillType(drillType.getMinorType());
+    if (sqlTypeName == SqlTypeName.DECIMAL) {
+      return typeFactory.createTypeWithNullability(
+          typeFactory.createSqlType(sqlTypeName, drillType.getPrecision(),
+              drillType.getScale()), isNullable);
+    }
+    return createCalciteTypeWithNullability(typeFactory, sqlTypeName, isNullable);
+  }
+
   /**
    * Given a SqlOperatorBinding, convert it to FunctionCall
    * @param  opBinding    the given SqlOperatorBinding
@@ -802,66 +997,28 @@ public class TypeInferenceUtils {
     for (int i = 0; i < opBinding.getOperandCount(); ++i) {
       final RelDataType type = opBinding.getOperandType(i);
       final TypeProtos.MinorType minorType = getDrillTypeFromCalciteType(type);
-      final TypeProtos.MajorType majorType;
-      if (type.isNullable()) {
-        majorType = Types.optional(minorType);
-      } else {
-        majorType = Types.required(minorType);
+      TypeProtos.DataMode dataMode =
+          type.isNullable() ? TypeProtos.DataMode.OPTIONAL : TypeProtos.DataMode.REQUIRED;
+
+      TypeProtos.MajorType.Builder builder =
+          TypeProtos.MajorType.newBuilder()
+              .setMode(dataMode)
+              .setMinorType(minorType);
+
+      if (Types.isDecimalType(minorType)) {
+        builder
+            .setScale(type.getScale())
+            .setPrecision(type.getPrecision());
       }
 
-      args.add(new MajorTypeInLogicalExpression(majorType));
+      args.add(new MajorTypeInLogicalExpression(builder.build()));
     }
 
     final String drillFuncName = FunctionCallFactory.replaceOpWithFuncName(opBinding.getOperator().getName());
-    final FunctionCall functionCall = new FunctionCall(
+    return new FunctionCall(
         drillFuncName,
         args,
         ExpressionPosition.UNKNOWN);
-    return functionCall;
-  }
-
-  public static SqlOperatorBinding convertDecimalLiteralToDouble(final SqlOperatorBinding sqlOperatorBinding) {
-    return new SqlOperatorBinding(sqlOperatorBinding.getTypeFactory(), sqlOperatorBinding.getOperator()) {
-      @Override
-      public int getOperandCount() {
-        return sqlOperatorBinding.getOperandCount();
-      }
-
-      @Override
-      public RelDataType getOperandType(int ordinal) {
-        final RelDataType relDataType;
-        if(isDecimalLiteral(sqlOperatorBinding, ordinal)) {
-          relDataType = createCalciteTypeWithNullability(
-              sqlOperatorBinding.getTypeFactory(),
-              SqlTypeName.DOUBLE,
-              sqlOperatorBinding.getOperandType(ordinal).isNullable());
-        } else {
-          relDataType = sqlOperatorBinding.getOperandType(ordinal);
-        }
-        return relDataType;
-      }
-
-      @Override
-      public CalciteException newError(Resources.ExInst<SqlValidatorException> e) {
-        throw new UnsupportedOperationException();
-      }
-    };
-  }
-
-  private static boolean isDecimalLiteral(SqlOperatorBinding sqlOperatorBinding, int index) {
-    if(sqlOperatorBinding.getOperandType(index).getSqlTypeName() != SqlTypeName.DECIMAL) {
-      return false;
-    }
-
-    final boolean isLiteral;
-    // TODO: The following try-catch statement is added due to one issue CALCITE-1176
-    // After that is fixed, the entire try-catch should be removed.
-    try {
-      isLiteral = sqlOperatorBinding.getOperandLiteralValue(index) != null;
-    } catch (AssertionError | UnsupportedOperationException exception) {
-      return false;
-    }
-    return isLiteral;
   }
 
   /**
@@ -885,4 +1042,4 @@ public class TypeInferenceUtils {
   private TypeInferenceUtils() {
 
   }
-}
+}
\ No newline at end of file
