@@ -17,20 +17,18 @@
  */
 package org.apache.drill.exec.store.mapr.db.json;
 
-import static org.apache.drill.exec.planner.index.Statistics.ROWCOUNT_HUGE;
-import static org.apache.drill.exec.planner.index.Statistics.ROWCOUNT_UNKNOWN;
-import static org.apache.drill.exec.planner.index.Statistics.AVG_ROWSIZE_UNKNOWN;
-
-import java.io.IOException;
-import java.util.HashMap;
-import java.util.List;
-import java.util.Map;
-import java.util.NavigableMap;
-import java.util.TreeMap;
-
+import com.fasterxml.jackson.annotation.JacksonInject;
+import com.fasterxml.jackson.annotation.JsonIgnore;
+import com.fasterxml.jackson.annotation.JsonProperty;
+import com.fasterxml.jackson.annotation.JsonTypeName;
+import org.apache.drill.shaded.guava.com.google.common.base.Preconditions;
+import com.mapr.db.MetaTable;
+import com.mapr.db.Table;
 import com.mapr.db.impl.ConditionImpl;
 import com.mapr.db.impl.ConditionNode.RowkeyRange;
-
+import com.mapr.db.impl.TabletInfoImpl;
+import com.mapr.db.index.IndexDesc;
+import com.mapr.db.scan.ScanRange;
 import org.apache.calcite.rel.RelNode;
 import org.apache.calcite.rex.RexNode;
 import org.apache.drill.common.exceptions.DrillRuntimeException;
@@ -43,17 +41,18 @@ import org.apache.drill.exec.physical.base.IndexGroupScan;
 import org.apache.drill.exec.physical.base.PhysicalOperator;
 import org.apache.drill.exec.physical.base.ScanStats;
 import org.apache.drill.exec.physical.base.ScanStats.GroupScanProperty;
-import org.apache.drill.exec.planner.index.Statistics;
+import org.apache.drill.exec.record.metadata.TupleSchema;
+import org.apache.drill.exec.planner.cost.PluginCost;
 import org.apache.drill.exec.planner.index.IndexDescriptor;
 import org.apache.drill.exec.planner.index.MapRDBIndexDescriptor;
 import org.apache.drill.exec.planner.index.MapRDBStatistics;
 import org.apache.drill.exec.planner.index.MapRDBStatisticsPayload;
-import org.apache.drill.exec.planner.cost.PluginCost;
+import org.apache.drill.exec.planner.index.Statistics;
 import org.apache.drill.exec.planner.physical.PartitionFunction;
 import org.apache.drill.exec.planner.physical.PrelUtil;
+import org.apache.drill.exec.store.AbstractStoragePlugin;
 import org.apache.drill.exec.store.StoragePluginRegistry;
 import org.apache.drill.exec.store.dfs.FileSystemConfig;
-import org.apache.drill.exec.store.dfs.FileSystemPlugin;
 import org.apache.drill.exec.store.mapr.PluginConstants;
 import org.apache.drill.exec.store.mapr.db.MapRDBFormatPlugin;
 import org.apache.drill.exec.store.mapr.db.MapRDBFormatPluginConfig;
@@ -63,19 +62,22 @@ import org.apache.drill.exec.store.mapr.db.MapRDBSubScanSpec;
 import org.apache.drill.exec.store.mapr.db.MapRDBTableStats;
 import org.apache.drill.exec.store.mapr.db.TabletFragmentInfo;
 import org.apache.drill.exec.util.Utilities;
+import org.apache.drill.exec.physical.base.FileSystemMetadataProviderManager;
+import org.apache.drill.exec.physical.base.MetadataProviderManager;
+import org.apache.drill.exec.physical.base.TableMetadataProvider;
 import org.codehaus.jackson.annotate.JsonCreator;
 import org.ojai.store.QueryCondition;
 
-import com.fasterxml.jackson.annotation.JacksonInject;
-import com.fasterxml.jackson.annotation.JsonIgnore;
-import com.fasterxml.jackson.annotation.JsonProperty;
-import com.fasterxml.jackson.annotation.JsonTypeName;
-import com.google.common.base.Preconditions;
-import com.mapr.db.MetaTable;
-import com.mapr.db.Table;
-import com.mapr.db.impl.TabletInfoImpl;
-import com.mapr.db.index.IndexDesc;
-import com.mapr.db.scan.ScanRange;
+import java.io.IOException;
+import java.util.HashMap;
+import java.util.List;
+import java.util.Map;
+import java.util.NavigableMap;
+import java.util.TreeMap;
+
+import static org.apache.drill.exec.planner.index.Statistics.AVG_ROWSIZE_UNKNOWN;
+import static org.apache.drill.exec.planner.index.Statistics.ROWCOUNT_HUGE;
+import static org.apache.drill.exec.planner.index.Statistics.ROWCOUNT_UNKNOWN;
 
 @JsonTypeName("maprdb-json-scan")
 public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupScan {
@@ -101,11 +103,12 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
   protected JsonScanSpec scanSpec;
   protected double fullTableRowCount;
   protected double fullTableEstimatedSize;
+  private boolean complexFilterCanBePushed = false;
 
   /**
    * need only read maxRecordsToRead records.
    */
-  protected long maxRecordsToRead = -1;
+  protected int maxRecordsToRead = -1;
 
   /**
    * Forced parallelization width
@@ -118,26 +121,25 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
                             @JsonProperty("storage") FileSystemConfig storagePluginConfig,
                             @JsonProperty("format") MapRDBFormatPluginConfig formatPluginConfig,
                             @JsonProperty("columns") List<SchemaPath> columns,
-                            @JacksonInject StoragePluginRegistry pluginRegistry) throws IOException, ExecutionSetupException {
-    this (userName,
-          (FileSystemPlugin) pluginRegistry.getPlugin(storagePluginConfig),
-          (MapRDBFormatPlugin) pluginRegistry.getFormatPlugin(storagePluginConfig, formatPluginConfig),
-          scanSpec, columns);
+                            // TODO: DRILL-7314 - replace TupleSchema with TupleMetadata
+                            @JsonProperty("schema") TupleSchema schema,
+                            @JacksonInject StoragePluginRegistry pluginRegistry) throws ExecutionSetupException, IOException {
+    this(userName, (AbstractStoragePlugin) pluginRegistry.getPlugin(storagePluginConfig),
+        (MapRDBFormatPlugin) pluginRegistry.getFormatPlugin(storagePluginConfig, formatPluginConfig),
+        scanSpec, columns, new MapRDBStatistics(), FileSystemMetadataProviderManager.getMetadataProviderForSchema(schema));
   }
 
-  public JsonTableGroupScan(String userName, FileSystemPlugin storagePlugin,
-                            MapRDBFormatPlugin formatPlugin, JsonScanSpec scanSpec, List<SchemaPath> columns) {
-    super(storagePlugin, formatPlugin, columns, userName);
-    this.scanSpec = scanSpec;
-    this.stats = new MapRDBStatistics();
-    this.forcedRowCountMap = new HashMap<>();
-    init();
-  }
-
-  public JsonTableGroupScan(String userName, FileSystemPlugin storagePlugin,
+  public JsonTableGroupScan(String userName, AbstractStoragePlugin storagePlugin,
                             MapRDBFormatPlugin formatPlugin, JsonScanSpec scanSpec, List<SchemaPath> columns,
-                            MapRDBStatistics stats) {
-    super(storagePlugin, formatPlugin, columns, userName);
+                            MetadataProviderManager metadataProviderManager) throws IOException {
+    this(userName, storagePlugin, formatPlugin, scanSpec, columns,
+        new MapRDBStatistics(), FileSystemMetadataProviderManager.getMetadataProvider(metadataProviderManager));
+  }
+
+  public JsonTableGroupScan(String userName, AbstractStoragePlugin storagePlugin,
+                            MapRDBFormatPlugin formatPlugin, JsonScanSpec scanSpec, List<SchemaPath> columns,
+                            MapRDBStatistics stats, TableMetadataProvider metadataProvider) {
+    super(storagePlugin, formatPlugin, columns, userName, metadataProvider);
     this.scanSpec = scanSpec;
     this.stats = stats;
     this.forcedRowCountMap = new HashMap<>();
@@ -193,7 +195,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
           PluginCost pluginCostModel = formatPlugin.getPluginCostModel();
           final int avgColumnSize = pluginCostModel.getAverageColumnSize(this);
           final int numColumns = (columns == null || columns.isEmpty() || Utilities.isStarQuery(columns)) ? STAR_COLS : columns.size();
-          MapRDBTableStats tableStats = new MapRDBTableStats(storagePlugin.getFsConf(), scanSpec.getTableName());
+          MapRDBTableStats tableStats = new MapRDBTableStats(storagePlugin.getConf(), scanSpec.getTableName());
           fullTableRowCount = tableStats.getNumRows();
           fullTableEstimatedSize = fullTableRowCount * numColumns * avgColumnSize;
         }
@@ -292,7 +294,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     assert minorFragmentId < endpointFragmentMapping.size() : String.format(
         "Mappings length [%d] should be greater than minor fragment id [%d] but it isn't.", endpointFragmentMapping.size(),
         minorFragmentId);
-    return new MapRDBSubScan(getUserName(), formatPlugin, endpointFragmentMapping.get(minorFragmentId), columns, maxRecordsToRead, TABLE_JSON);
+    return new MapRDBSubScan(getUserName(), formatPlugin, endpointFragmentMapping.get(minorFragmentId), columns, maxRecordsToRead, TABLE_JSON, getSchema());
   }
 
   @Override
@@ -300,7 +302,6 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     if (isIndexScan()) {
       return indexScanStats();
     }
-
     return fullTableScanStats();
   }
 
@@ -318,7 +319,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     double totalRowCount = stats.getRowCount(null, null);
     logger.debug("GroupScan {} with stats {}: rowCount={}, condition={}, totalRowCount={}, fullTableRowCount={}",
             System.identityHashCode(this), System.identityHashCode(stats), rowCount,
-            scanSpec.getCondition()==null?"null":scanSpec.getCondition(),
+            scanSpec.getCondition() == null ? "null" : scanSpec.getCondition(),
             totalRowCount, fullTableRowCount);
     // If UNKNOWN, or DB stats sync issues(manifests as 0 rows) use defaults.
     if (rowCount == ROWCOUNT_UNKNOWN || rowCount == 0) {
@@ -366,7 +367,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
 
   private ScanStats indexScanStats() {
     if (!this.getIndexHint().equals("") &&
-         this.getIndexHint().equals(getIndexDesc().getIndexName())) {
+        this.getIndexHint().equals(getIndexDesc().getIndexName())) {
       logger.debug("JsonIndexGroupScan:{} forcing index {} by making tiny cost", this, this.getIndexHint());
       return new ScanStats(GroupScanProperty.NO_EXACT_ROW_COUNT, 1,1, 0);
     }
@@ -375,7 +376,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     PluginCost pluginCostModel = formatPlugin.getPluginCostModel();
     final int avgColumnSize = pluginCostModel.getAverageColumnSize(this);
     boolean filterPushed = (scanSpec.getSerializedFilter() != null);
-    if(scanSpec != null && scanSpec.getIndexDesc() != null) {
+    if (scanSpec != null && scanSpec.getIndexDesc() != null) {
       totalColNum = scanSpec.getIndexDesc().getIncludedFields().size()
           + scanSpec.getIndexDesc().getIndexedFields().size() + 1;
     }
@@ -419,7 +420,9 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
   @JsonIgnore
   public PhysicalOperator getNewWithChildren(List<PhysicalOperator> children) {
     Preconditions.checkArgument(children.isEmpty());
-    return new JsonTableGroupScan(this);
+    JsonTableGroupScan groupScan = new JsonTableGroupScan(this);
+    groupScan.setComplexFilterPushDown(this.complexFilterCanBePushed);
+    return groupScan;
   }
 
   @Override
@@ -444,8 +447,8 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
   @Override
   public String toString() {
     return "JsonTableGroupScan [ScanSpec=" + scanSpec + ", columns=" + columns
-        + (maxRecordsToRead>0? ", limit=" + maxRecordsToRead : "")
-        + (getMaxParallelizationWidth()>0? ", maxwidth=" + getMaxParallelizationWidth() : "") + "]";
+        + (maxRecordsToRead > 0 ? ", limit=" + maxRecordsToRead : "")
+        + (getMaxParallelizationWidth() > 0 ? ", maxwidth=" + getMaxParallelizationWidth() : "") + "]";
   }
 
   public JsonScanSpec getScanSpec() {
@@ -468,17 +471,24 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     return true;
   }
 
+
   @Override
   public RestrictedJsonTableGroupScan getRestrictedScan(List<SchemaPath> columns) {
-    RestrictedJsonTableGroupScan newScan =
-        new RestrictedJsonTableGroupScan(this.getUserName(),
+    try {
+      RestrictedJsonTableGroupScan newScan = new RestrictedJsonTableGroupScan(this.getUserName(),
             this.getStoragePlugin(),
             this.getFormatPlugin(),
             this.getScanSpec(),
             this.getColumns(),
-            this.getStatistics());
-    newScan.columns = columns;
-    return newScan;
+            this.getStatistics(),
+            // TODO: DRILL-7314 - replace TupleSchema with TupleMetadata
+            (TupleSchema) this.getSchema());
+      newScan.setComplexFilterPushDown(this.complexFilterCanBePushed);
+      newScan.columns = columns;
+      return newScan;
+    } catch (IOException e) {
+      throw new DrillRuntimeException("Error happened when constructing RestrictedJsonTableGroupScan", e);
+    }
   }
 
   /**
@@ -487,8 +497,9 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
    * @param index, to use for generating the estimate
    * @return row count post filtering
    */
-  public MapRDBStatisticsPayload getAverageRowSizeStats(IndexDescriptor index) {
+  public MapRDBStatisticsPayload getFilterIndependentStats(IndexDescriptor index) {
     IndexDesc indexDesc = null;
+    double totalRowCount = ROWCOUNT_UNKNOWN;
     double avgRowSize = AVG_ROWSIZE_UNKNOWN;
 
     if (index != null) {
@@ -496,7 +507,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     }
     // If no index is specified, get it from the primary table
     if(indexDesc == null && scanSpec.isSecondaryIndex()) {
-      throw new UnsupportedOperationException("getAverageRowSizeStats should be invoked on primary table");
+      throw new UnsupportedOperationException("getFilterIndependentStats should be invoked on primary table");
     }
 
     // Get the index table or primary table and use the DB API to get the estimated number of rows. For size estimates,
@@ -505,14 +516,16 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
 
     if (table != null) {
       final MetaTable metaTable = table.getMetaTable();
+      final com.mapr.db.scan.ScanStats stats = metaTable.getScanStats();
       if (metaTable != null) {
         avgRowSize = metaTable.getAverageRowSize();
+        totalRowCount = stats.getEstimatedNumRows();
       }
     }
     logger.debug("index_plan_info: getEstimatedRowCount obtained from DB Client for {}: indexName: {}, indexInfo: {}, " +
-        "avgRowSize: {}, estimatedSize {}", this, (indexDesc == null ? "null" : indexDesc.getIndexName()),
-        (indexDesc == null ? "null" : indexDesc.getIndexInfo()), avgRowSize);
-    return new MapRDBStatisticsPayload(ROWCOUNT_UNKNOWN, ROWCOUNT_UNKNOWN, avgRowSize);
+            " avgRowSize {}, totalRowCount: {}", this, (indexDesc == null ? "null" : indexDesc.getIndexName()),
+        (indexDesc == null ? "null" : indexDesc.getIndexInfo()), avgRowSize, totalRowCount);
+    return new MapRDBStatisticsPayload(totalRowCount, ROWCOUNT_UNKNOWN, avgRowSize);
   }
 
   /**
@@ -537,12 +550,11 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
    * @return {@link MapRDBStatisticsPayload} statistics
    */
   private MapRDBStatisticsPayload getFirstKeyEstimatedStatsInternal(QueryCondition condition, IndexDesc index, RelNode scanRel) {
-    // double totalRows = getRowCount(null, scanPrel);
 
     // If no index is specified, get it from the primary table
-    if(index == null && scanSpec.isSecondaryIndex()) {
+    if (index == null && scanSpec.isSecondaryIndex()) {
       // If stats not cached get it from the table.
-      //table = MapRDB.getTable(scanSpec.getPrimaryTablePath());
+      // table = MapRDB.getTable(scanSpec.getPrimaryTablePath());
       throw new UnsupportedOperationException("getFirstKeyEstimatedStats should be invoked on primary table");
     }
 
@@ -587,7 +599,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
       }
       logger.info("index_plan_info: getEstimatedRowCount obtained from DB Client for {}: indexName: {}, indexInfo: {}, " +
               "condition: {} rowCount: {}, avgRowSize: {}, estimatedSize {}, tabletCount {}, totalTabletCount {}, " +
-          "scalingFactor {}",
+              "scalingFactor {}",
           this, (index == null ? "null" : index.getIndexName()), (index == null ? "null" : index.getIndexInfo()),
           (condition == null ? "null" : condition.toString()), stats.getEstimatedNumRows(),
           (stats.getEstimatedNumRows() == 0 ? 0 : stats.getEstimatedSize()/stats.getEstimatedNumRows()),
@@ -602,6 +614,7 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     }
   }
 
+
   /**
    * Set the row count resulting from applying the {@link RexNode} condition. Forced row counts will take
    * precedence over stats row counts
@@ -736,11 +749,11 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
     if (maxRecordsToRead < 0) {
       return true;
     }
-    return false;//limit is already pushed. No more pushdown of limit
+    return false; // limit is already pushed. No more pushdown of limit
   }
 
   @Override
-  public GroupScan applyLimit(long maxRecords) {
+  public GroupScan applyLimit(int maxRecords) {
     maxRecordsToRead = Math.max(maxRecords, 1);
     return this;
   }
@@ -760,4 +773,14 @@ public class JsonTableGroupScan extends MapRDBGroupScan implements IndexGroupSca
       logger.debug("Forced parallelization width = {}", width);
     }
   }
+
+  @Override
+  @JsonIgnore
+  public void setComplexFilterPushDown(boolean complexFilterPushDown) {
+    complexFilterCanBePushed = complexFilterPushDown;
+  }
+
+  @Override
+  @JsonIgnore
+  public boolean supportsComplexFilterPushDown() { return complexFilterCanBePushed; }
 }
