@@ -17,39 +17,38 @@
  */
 package org.apache.drill.exec.store.mapr.db.json;
 
-import static org.apache.drill.exec.store.mapr.PluginConstants.DOCUMENT_SCHEMA_PATH;
-import static org.apache.drill.exec.store.mapr.PluginErrorHandler.dataReadError;
-import static org.ojai.DocumentConstants.ID_FIELD;
-
-import java.util.Collection;
-import java.util.Iterator;
-import java.util.List;
-import java.util.Set;
-import java.util.Stack;
-import java.util.Collections;
-import java.util.concurrent.TimeUnit;
 
+import com.mapr.db.Table;
+import com.mapr.db.Table.TableOption;
+import com.mapr.db.exceptions.DBException;
+import com.mapr.db.impl.IdCodec;
 import com.mapr.db.impl.MapRDBImpl;
+import com.mapr.db.index.IndexDesc;
+import com.mapr.db.ojai.DBDocumentReaderBase;
+import com.mapr.db.util.ByteBufs;
+import io.netty.buffer.DrillBuf;
 import org.apache.drill.common.exceptions.ExecutionSetupException;
 import org.apache.drill.common.exceptions.UserException;
 import org.apache.drill.common.expression.PathSegment;
 import org.apache.drill.common.expression.SchemaPath;
 import org.apache.drill.exec.ExecConstants;
 import org.apache.drill.exec.exception.SchemaChangeException;
+import org.apache.drill.exec.expr.fn.impl.DateUtility;
 import org.apache.drill.exec.ops.FragmentContext;
 import org.apache.drill.exec.ops.OperatorContext;
 import org.apache.drill.exec.ops.OperatorStats;
 import org.apache.drill.exec.physical.impl.OutputMutator;
+import org.apache.drill.exec.record.metadata.TupleMetadata;
 import org.apache.drill.exec.store.AbstractRecordReader;
 import org.apache.drill.exec.store.mapr.db.MapRDBFormatPlugin;
 import org.apache.drill.exec.store.mapr.db.MapRDBSubScanSpec;
 import org.apache.drill.exec.util.EncodedSchemaPathSet;
-import org.apache.drill.exec.util.Utilities;
 import org.apache.drill.exec.vector.BaseValueVector;
-import org.apache.drill.exec.vector.complex.impl.MapOrListWriterImpl;
 import org.apache.drill.exec.vector.complex.fn.JsonReaderUtils;
+import org.apache.drill.exec.vector.complex.impl.MapOrListWriterImpl;
 import org.apache.drill.exec.vector.complex.impl.VectorContainerWriter;
 import org.apache.hadoop.fs.Path;
+import org.ojai.Document;
 import org.ojai.DocumentReader;
 import org.ojai.DocumentStream;
 import org.ojai.FieldPath;
@@ -59,24 +58,30 @@ import org.ojai.util.FieldProjector;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
-import com.google.common.base.Preconditions;
-import com.google.common.base.Predicate;
-import com.google.common.base.Stopwatch;
-import com.google.common.collect.ImmutableSet;
-import com.google.common.collect.Iterables;
-import com.google.common.collect.Sets;
-import com.mapr.db.Table;
-import com.mapr.db.Table.TableOption;
-import com.mapr.db.exceptions.DBException;
-import com.mapr.db.impl.IdCodec;
-import com.mapr.db.index.IndexDesc;
-import com.mapr.db.ojai.DBDocumentReaderBase;
-import com.mapr.db.util.ByteBufs;
+import org.apache.drill.shaded.guava.com.google.common.base.Preconditions;
+import org.apache.drill.shaded.guava.com.google.common.base.Stopwatch;
+import org.apache.drill.shaded.guava.com.google.common.collect.ImmutableSet;
+import org.apache.drill.shaded.guava.com.google.common.collect.Iterables;
+import org.apache.drill.shaded.guava.com.google.common.collect.Sets;
 
-import io.netty.buffer.DrillBuf;
+import java.time.Instant;
+import java.time.ZoneId;
+import java.time.ZonedDateTime;
+import java.util.Collection;
+import java.util.Collections;
+import java.util.Iterator;
+import java.util.List;
+import java.util.Set;
+import java.util.Stack;
+import java.util.concurrent.TimeUnit;
+
+import static org.apache.drill.exec.store.mapr.PluginConstants.DOCUMENT_SCHEMA_PATH;
+import static org.apache.drill.exec.store.mapr.PluginErrorHandler.dataReadError;
+import static org.ojai.DocumentConstants.ID_FIELD;
 
 public class MaprDBJsonRecordReader extends AbstractRecordReader {
   private static final Logger logger = LoggerFactory.getLogger(MaprDBJsonRecordReader.class);
+  protected enum SchemaState {SCHEMA_UNKNOWN, SCHEMA_INIT, SCHEMA_CHANGE};
 
   protected static final FieldPath[] ID_ONLY_PROJECTION = { ID_FIELD };
 
@@ -91,55 +96,62 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
    */
   private FieldPath[] scannedFields;
 
-  private final Path tablePath;
-  private final IndexDesc indexDesc;
   private OperatorContext operatorContext;
   protected VectorContainerWriter vectorWriter;
+  private DBDocumentReaderBase reader;
+  Document document;
+  protected OutputMutator vectorWriterMutator;
 
   private DrillBuf buffer;
 
   private DocumentStream documentStream;
 
   private Iterator<DocumentReader> documentReaderIterators;
+  private Iterator<Document> documentIterator;
 
   private boolean includeId;
   private boolean idOnly;
-
+  private SchemaState schemaState;
   private boolean projectWholeDocument;
   private FieldProjector projector;
 
   private final boolean unionEnabled;
   private final boolean readNumbersAsDouble;
+  private final boolean readTimestampWithZoneOffset;
   private boolean disablePushdown;
   private final boolean allTextMode;
   private final boolean ignoreSchemaChange;
   private final boolean disableCountOptimization;
   private final boolean nonExistentColumnsProjection;
+  private final TupleMetadata schema;
 
   protected final MapRDBSubScanSpec subScanSpec;
   protected final MapRDBFormatPlugin formatPlugin;
 
   protected OjaiValueWriter valueWriter;
   protected DocumentReaderVectorWriter documentWriter;
-  protected long maxRecordsToRead = -1;
+  protected int maxRecordsToRead = -1;
+  protected DBDocumentReaderBase lastDocumentReader;
+  protected Document lastDocument;
 
   public MaprDBJsonRecordReader(MapRDBSubScanSpec subScanSpec, MapRDBFormatPlugin formatPlugin,
-                                List<SchemaPath> projectedColumns, FragmentContext context, long maxRecords) {
-    this(subScanSpec, formatPlugin, projectedColumns, context);
+                                List<SchemaPath> projectedColumns, FragmentContext context, int maxRecords, TupleMetadata schema) {
+    this(subScanSpec, formatPlugin, projectedColumns, context, schema);
     this.maxRecordsToRead = maxRecords;
+    this.lastDocumentReader = null;
+    this.lastDocument = null;
+    this.schemaState = SchemaState.SCHEMA_UNKNOWN;
   }
 
-  public MaprDBJsonRecordReader(MapRDBSubScanSpec subScanSpec, MapRDBFormatPlugin formatPlugin,
-                                List<SchemaPath> projectedColumns, FragmentContext context) {
+  protected MaprDBJsonRecordReader(MapRDBSubScanSpec subScanSpec, MapRDBFormatPlugin formatPlugin,
+                                List<SchemaPath> projectedColumns, FragmentContext context, TupleMetadata schema) {
     buffer = context.getManagedBuffer();
-    scannedFields = null;
-    tablePath = new Path(Preconditions.checkNotNull(subScanSpec, "MapRDB reader needs a sub-scan spec").getTableName());
+    final Path tablePath = new Path(Preconditions.checkNotNull(subScanSpec,
+      "MapRDB reader needs a sub-scan spec").getTableName());
     this.subScanSpec = subScanSpec;
     this.formatPlugin = formatPlugin;
-    indexDesc = subScanSpec.getIndexDesc();
-    projectWholeDocument = false;
-    includeId = false;
-    idOnly    = false;
+    this.schema = schema;
+    final IndexDesc indexDesc = subScanSpec.getIndexDesc();
     byte[] serializedFilter = subScanSpec.getSerializedFilter();
     condition = null;
 
@@ -148,14 +160,26 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
     }
 
     disableCountOptimization = formatPlugin.getConfig().disableCountOptimization();
+    // Below call will set the scannedFields and includeId correctly
     setColumns(projectedColumns);
     unionEnabled = context.getOptions().getOption(ExecConstants.ENABLE_UNION_TYPE);
     readNumbersAsDouble = formatPlugin.getConfig().isReadAllNumbersAsDouble();
+    readTimestampWithZoneOffset = formatPlugin.getConfig().isReadTimestampWithZoneOffset();
     allTextMode = formatPlugin.getConfig().isAllTextMode();
     ignoreSchemaChange = formatPlugin.getConfig().isIgnoreSchemaChange();
     disablePushdown = !formatPlugin.getConfig().isEnablePushdown();
     nonExistentColumnsProjection = formatPlugin.getConfig().isNonExistentFieldSupport();
 
+    // Do not use cached table handle for two reasons.
+    // cached table handles default timeout is 60 min after which those handles will become stale.
+    // Since execution can run for longer than 60 min, we want to get a new table handle and use it
+    // instead of the one from cache.
+    // Since we are setting some table options, we do not want to use shared handles.
+    //
+    // Call it here instead of setup since this will make sure it's called under correct UGI block when impersonation
+    // is enabled and table is used with and without views.
+    table = (indexDesc == null ? MapRDBImpl.getTable(tablePath) : MapRDBImpl.getIndexTable(indexDesc));
+
     if (condition != null) {
       logger.debug("Created record reader with query condition {}", condition.toString());
     } else {
@@ -169,11 +193,11 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
     Set<SchemaPath> encodedSchemaPathSet = Sets.newLinkedHashSet();
 
     if (disablePushdown) {
-      transformed.add(Utilities.STAR_COLUMN);
+      transformed.add(SchemaPath.STAR_COLUMN);
       includeId = true;
     } else {
       if (isStarQuery()) {
-        transformed.add(Utilities.STAR_COLUMN);
+        transformed.add(SchemaPath.STAR_COLUMN);
         includeId = true;
         if (isSkipQuery() && !disableCountOptimization) {
           // `SELECT COUNT(*)` query
@@ -209,7 +233,7 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
           // now we look at the fields which are part of encoded field set and either
           // add them to scanned set or clear the scanned set if all fields were requested.
           for (SchemaPath column : decodedSchemaPaths) {
-            if (column.equals(Utilities.STAR_COLUMN)) {
+            if (column.equals(SchemaPath.STAR_COLUMN)) {
               includeId = true;
               scannedFieldsSet.clear();
               break;
@@ -229,7 +253,7 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
           idOnly = (scannedFields == null);
         }
 
-        if(projectWholeDocument) {
+        if (projectWholeDocument) {
           projector = new FieldProjector(projectedFieldsSet);
         }
 
@@ -257,45 +281,91 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
   @Override
   public void setup(OperatorContext context, OutputMutator output) throws ExecutionSetupException {
     this.vectorWriter = new VectorContainerWriter(output, unionEnabled);
+    this.vectorWriterMutator = output;
     this.operatorContext = context;
 
     try {
-      // Do not use cached table handle for two reasons.
-      // cached table handles default timeout is 60 min after which those handles will become stale.
-      // Since execution can run for longer than 60 min, we want to get a new table handle and use it
-      // instead of the one from cache.
-      // Since we are setting some table options, we do not want to use shared handles.
-
-      // If for some reason, table is not null (may happen if setup gets called multiple times ?),
-      // close the table.
-      if (table != null) {
-        table.close();
-      }
-      table = (indexDesc == null ? MapRDBImpl.getTable(tablePath) : MapRDBImpl.getIndexTable(indexDesc));
       table.setOption(TableOption.EXCLUDEID, !includeId);
       documentStream = table.find(condition, scannedFields);
-      documentReaderIterators = documentStream.documentReaders().iterator();
+      documentIterator = documentStream.iterator();
+      setupWriter();
+    } catch (DBException ex) {
+      throw new ExecutionSetupException(ex);
+    }
+  }
 
-      if (allTextMode) {
+  /**
+   * Setup the valueWriter and documentWriters based on config options.
+   */
+  private void setupWriter() {
+    if (allTextMode) {
+      if (readTimestampWithZoneOffset) {
+        valueWriter = new AllTextValueWriter(buffer) {
+          /**
+           * Applies local time zone offset to timestamp value read using specified {@code reader}.
+           *
+           * @param writer    writer to store string representation of timestamp value
+           * @param fieldName name of the field
+           * @param reader    document reader
+           */
+          @Override
+          protected void writeTimeStamp(MapOrListWriterImpl writer, String fieldName, DocumentReader reader) {
+            String formattedTimestamp = Instant.ofEpochMilli(reader.getTimestampLong())
+                .atZone(ZoneId.systemDefault()).format(DateUtility.UTC_FORMATTER);
+            writeString(writer, fieldName, formattedTimestamp);
+          }
+        };
+      } else {
         valueWriter = new AllTextValueWriter(buffer);
-      } else if (readNumbersAsDouble) {
+      }
+    } else if (readNumbersAsDouble) {
+      if (readTimestampWithZoneOffset) {
+        valueWriter = new NumbersAsDoubleValueWriter(buffer) {
+          @Override
+          protected void writeTimeStamp(MapOrListWriterImpl writer, String fieldName, DocumentReader reader) {
+            writeTimestampWithLocalZoneOffset(writer, fieldName, reader);
+          }
+        };
+      } else {
         valueWriter = new NumbersAsDoubleValueWriter(buffer);
+      }
+    } else {
+      if (readTimestampWithZoneOffset) {
+        valueWriter = new OjaiValueWriter(buffer) {
+          @Override
+          protected void writeTimeStamp(MapOrListWriterImpl writer, String fieldName, DocumentReader reader) {
+            writeTimestampWithLocalZoneOffset(writer, fieldName, reader);
+          }
+        };
       } else {
         valueWriter = new OjaiValueWriter(buffer);
       }
-
-      if (projectWholeDocument) {
-        documentWriter = new ProjectionPassthroughVectorWriter(valueWriter, projector, includeId);
-      } else if (isSkipQuery()) {
-        documentWriter = new RowCountVectorWriter(valueWriter);
-      } else if (idOnly) {
-        documentWriter = new IdOnlyVectorWriter(valueWriter);
-      } else {
-        documentWriter = new FieldTransferVectorWriter(valueWriter);
-      }
-    } catch (DBException e) {
-      throw new ExecutionSetupException(e);
     }
+
+    if (projectWholeDocument) {
+      documentWriter = new ProjectionPassthroughVectorWriter(valueWriter, projector, includeId);
+    } else if (isSkipQuery()) {
+      documentWriter = new RowCountVectorWriter(valueWriter);
+    } else if (idOnly) {
+      documentWriter = new IdOnlyVectorWriter(valueWriter);
+    } else {
+      documentWriter = new FieldTransferVectorWriter(valueWriter);
+    }
+  }
+
+  /**
+   * Applies local time zone offset to timestamp value read using specified {@code reader}.
+   *
+   * @param writer    writer to store timestamp value
+   * @param fieldName name of the field
+   * @param reader    document reader
+   */
+  private void writeTimestampWithLocalZoneOffset(MapOrListWriterImpl writer, String fieldName, DocumentReader reader) {
+    Instant utcInstant = Instant.ofEpochMilli(reader.getTimestampLong());
+    ZonedDateTime localZonedDateTime = utcInstant.atZone(ZoneId.systemDefault());
+    ZonedDateTime convertedZonedDateTime = localZonedDateTime.withZoneSameLocal(ZoneId.of("UTC"));
+    long timeStamp = convertedZonedDateTime.toInstant().toEpochMilli();
+    writer.timeStamp(fieldName).writeTimeStamp(timeStamp);
   }
 
   @Override
@@ -307,40 +377,82 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
     vectorWriter.reset();
 
     int recordCount = 0;
-    DBDocumentReaderBase reader = null;
+    reader = null;
+    document = null;
 
-    long maxRecordsForThisBatch = this.maxRecordsToRead >= 0?
+    int maxRecordsForThisBatch = this.maxRecordsToRead >= 0?
         Math.min(BaseValueVector.INITIAL_VALUE_ALLOCATION, this.maxRecordsToRead) : BaseValueVector.INITIAL_VALUE_ALLOCATION;
 
+    try {
+      // If the last document caused a SchemaChange create a new output schema for this scan batch
+      if (schemaState == SchemaState.SCHEMA_CHANGE && !ignoreSchemaChange) {
+        // Clear the ScanBatch vector container writer/mutator in order to be able to generate the new schema
+        vectorWriterMutator.clear();
+        vectorWriter = new VectorContainerWriter(vectorWriterMutator, unionEnabled);
+        logger.debug("Encountered schema change earlier use new writer {}", vectorWriter.toString());
+        document = lastDocument;
+        setupWriter();
+        if (recordCount < maxRecordsForThisBatch) {
+          vectorWriter.setPosition(recordCount);
+          if (document != null) {
+            reader = (DBDocumentReaderBase) document.asReader();
+            documentWriter.writeDBDocument(vectorWriter, reader);
+            recordCount++;
+          }
+        }
+      }
+    } catch (SchemaChangeException e) {
+      String err_row = reader.getId().asJsonString();
+      if (ignoreSchemaChange) {
+        logger.warn("{}. Dropping row '{}' from result.", e.getMessage(), err_row);
+        logger.debug("Stack trace:", e);
+      } else {
+          /* We should not encounter a SchemaChangeException here since this is the first document for this
+           * new schema. Something is very wrong - cannot handle any further!
+           */
+        throw dataReadError(logger, e, "SchemaChangeException for row '%s'.", err_row);
+      }
+    }
+    schemaState = SchemaState.SCHEMA_INIT;
     while(recordCount < maxRecordsForThisBatch) {
       vectorWriter.setPosition(recordCount);
       try {
-        reader = nextDocumentReader();
-        if (reader == null) {
+        document = nextDocument();
+        if (document == null) {
           break; // no more documents for this reader
         } else {
-          documentWriter.writeDBDocument(vectorWriter, reader);
+          documentWriter.writeDBDocument(vectorWriter, (DBDocumentReaderBase) document.asReader());
         }
         recordCount++;
       } catch (UserException e) {
         throw UserException.unsupportedError(e)
             .addContext(String.format("Table: %s, document id: '%s'",
                 table.getPath(),
-                reader == null ? null : IdCodec.asString(reader.getId())))
+                    document.asReader() == null ? null :
+                        IdCodec.asString(((DBDocumentReaderBase)document.asReader()).getId())))
             .build(logger);
       } catch (SchemaChangeException e) {
-        String err_row = reader.getId().asJsonString();
+        String err_row = ((DBDocumentReaderBase)document.asReader()).getId().asJsonString();
         if (ignoreSchemaChange) {
           logger.warn("{}. Dropping row '{}' from result.", e.getMessage(), err_row);
           logger.debug("Stack trace:", e);
         } else {
-          throw dataReadError(logger, e, "SchemaChangeException for row '%s'.", err_row);
+          /* Save the current document reader for next iteration. The recordCount is not updated so we
+           * would start from this reader on the next next() call
+           */
+          lastDocument = document;
+          schemaState = SchemaState.SCHEMA_CHANGE;
+          break;
         }
       }
     }
 
     if (nonExistentColumnsProjection && recordCount > 0) {
-      JsonReaderUtils.ensureAtLeastOneField(vectorWriter, getColumns(), allTextMode, Collections.EMPTY_LIST);
+      if (schema == null || schema.isEmpty()) {
+        JsonReaderUtils.ensureAtLeastOneField(vectorWriter, getColumns(), allTextMode, Collections.emptyList());
+      } else {
+        JsonReaderUtils.writeColumnsUsingSchema(vectorWriter, getColumns(), schema, allTextMode);
+      }
     }
     vectorWriter.setValueCount(recordCount);
     if (maxRecordsToRead > 0) {
@@ -372,6 +484,27 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
     }
   }
 
+  protected Document nextDocument() {
+    final OperatorStats operatorStats = operatorContext == null ? null : operatorContext.getStats();
+    try {
+      if (operatorStats != null) {
+        operatorStats.startWait();
+      }
+      try {
+        if (!documentIterator.hasNext()) {
+          return null;
+        } else {
+          return documentIterator.next();
+        }
+      } finally {
+        if (operatorStats != null) {
+          operatorStats.stopWait();
+        }
+      }
+    } catch (DBException e) {
+      throw dataReadError(logger, e);
+    }
+  }
   /*
    * Extracts contiguous named segments from the SchemaPath, starting from the
    * root segment and build the FieldPath from it for projection.
@@ -395,12 +528,7 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
   }
 
   public static boolean includesIdField(Collection<FieldPath> projected) {
-    return Iterables.tryFind(projected, new Predicate<FieldPath>() {
-      @Override
-      public boolean apply(FieldPath path) {
-        return Preconditions.checkNotNull(path).equals(ID_FIELD);
-      }
-    }).isPresent();
+    return Iterables.tryFind(projected, path -> Preconditions.checkNotNull(path).equals(ID_FIELD)).isPresent();
   }
 
   @Override
@@ -408,11 +536,22 @@ public class MaprDBJsonRecordReader extends AbstractRecordReader {
     if (documentStream != null) {
       documentStream.close();
     }
-
     if (table != null) {
       table.close();
     }
-
   }
 
+  @Override
+  public String toString() {
+    StringBuilder sb = new StringBuilder("MaprDBJsonRecordReader[Table=")
+        .append(table != null ? table.getPath() : null);
+    if (reader != null) {
+      sb.append(", Document ID=")
+          .append(IdCodec.asString(reader.getId()));
+    }
+    sb.append(", reader=")
+        .append(reader)
+        .append(']');
+    return sb.toString();
+  }
 }
